/* Copyright Airship and Contributors */

#import "UABaseTest.h"
#import "UAChannelRegistrationPayload+Internal.h"
#import "NSJSONSerialization+UAAdditions.h"
#import "UAPush+Internal.h"
#import "UAChannelRegistrationPayload+Internal.h"
#import "UAAnalytics.h"
#import "UAirship+Internal.h"

@interface UAChannelRegistrationPayloadTest : UABaseTest
@property (nonatomic, strong) UAChannelRegistrationPayload *payload;

@property (nonatomic, strong) id mockAirship;
@property (nonatomic, strong) id mockAnalytics;

@end

@implementation UAChannelRegistrationPayloadTest

- (void)setUp {
    [super setUp];

    NSDictionary *quietTime = [self buildQuietTimeWithStartDate:[NSDate dateWithTimeIntervalSince1970:30]
                                                    withEndDate:[NSDate dateWithTimeIntervalSince1970:100]];

    self.mockAnalytics = [self mockForClass:[UAAnalytics class]];

    self.mockAirship = [self mockForClass:[UAirship class]];
    [UAirship setSharedAirship:self.mockAirship];
    [[[self.mockAirship stub] andReturn:self.mockAnalytics] analytics];

    self.payload = [[UAChannelRegistrationPayload alloc] init];

    // set up the full payload
    self.payload.optedIn = YES;
    self.payload.backgroundEnabled = YES;
    self.payload.pushAddress = @"FAKEADDRESS";
    self.payload.userID = @"fakeUser";
    self.payload.deviceID = @"fakeDeviceID";
    self.payload.accengageDeviceID = @"fakeAccengageDeviceID";
    self.payload.namedUserId = @"fakeNamedUserID";
    self.payload.badge = [NSNumber numberWithInteger:1];
    self.payload.quietTime =  quietTime;
    self.payload.timeZone = @"timezone";
    self.payload.quietTimeTimeZone = @"quietTimeTimeZone";
    self.payload.language = @"language";
    self.payload.country = @"country";
    self.payload.tags = @[@"tagOne", @"tagTwo"];
    self.payload.setTags = YES;
    self.payload.locationSettings = @(YES);
    self.payload.SDKVersion = @"SDKVersion";
    self.payload.appVersion = @"appVersion";
    self.payload.deviceModel = @"deviceModel";
    self.payload.deviceOS = @"deviceOS";
    self.payload.carrier = @"carrier";

}

- (void)tearDown {
    [self.mockAirship stopMocking];
    [self.mockAnalytics stopMocking];
}

/**
 * Test that the json has the full expected payload
 */
- (void)testAsJsonFullPayload {
    [[[self.mockAnalytics stub] andReturnValue:OCMOCK_VALUE(YES)] isEnabled];

    NSString *jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    NSDictionary *dict = [NSJSONSerialization objectWithString:jsonString];

    // identity hints
    NSDictionary *identityHints = [dict valueForKey:UAChannelIdentityHintsKey];
    XCTAssertNotNil(identityHints, @"identity hints should be present");
    XCTAssertEqualObjects(self.payload.userID, [identityHints valueForKey:UAChannelUserIDKey], @"user ID should be present");
    XCTAssertEqualObjects(self.payload.deviceID, [identityHints valueForKey:UAChannelDeviceIDKey], @"device ID should be present");

    // channel specific items
    NSDictionary *channel = [dict valueForKey:UAChannelKey];
    XCTAssertEqualObjects(@"ios", [channel valueForKey:UAChannelDeviceTypeKey], @"device type should be present");
    XCTAssertEqualObjects([NSNumber numberWithBool:self.payload.optedIn], [channel valueForKey:UAChannelOptInKey], @"opt-in should be present");
    XCTAssertEqualObjects([NSNumber numberWithBool:self.payload.backgroundEnabled], [channel valueForKey:UABackgroundEnabledJSONKey], @"background should be present");
    XCTAssertEqualObjects(self.payload.pushAddress, [channel valueForKey:UAChannelPushAddressKey], @"push address should be present");
    XCTAssertEqualObjects([NSNumber numberWithBool:self.payload.setTags], [channel valueForKey:UAChannelSetTagsKey], @"set tags should be present");
    XCTAssertEqualObjects(self.payload.tags, [channel valueForKey:UAChannelTagsJSONKey], @"tags should be present");

    // channel specific items
    XCTAssertEqualObjects(@"timezone", [channel valueForKey:UAChannelTopLevelTimeZoneJSONKey], @"timezone key should be available in the channel dictionary");
    XCTAssertEqualObjects(@"language", [channel valueForKey:UAChannelTopLevelLanguageJSONKey], @"locale_language key should be available in the channel dictionary");
    XCTAssertEqualObjects(@"country", [channel valueForKey:UAChannelTopLevelCountryJSONKey], @"locale_country key should be available in the channel dictionary");
    XCTAssertEqualObjects(@(YES), [channel valueForKey:UAChannelTopLevelLocationSettingsJSONKey]);
    XCTAssertEqualObjects(@"SDKVersion", [channel valueForKey:UAChannelTopLevelSDKVersionJSONKey]);
    XCTAssertEqualObjects(@"appVersion", [channel valueForKey:UAChannelTopLevelAppVersionJSONKey]);
    XCTAssertEqualObjects(@"deviceModel", [channel valueForKey:UAChannelTopLevelDeviceModelJSONKey]);
    XCTAssertEqualObjects(@"deviceOS", [channel valueForKey:UAChannelTopLevelDeviceOSJSONKey]);
    XCTAssertEqualObjects(@"carrier", [channel valueForKey:UAChannelTopLevelCarrierJSONKey]);

    // iOS specific items
    NSDictionary *ios = [channel valueForKey:UAChannelIOSKey];
    XCTAssertNotNil(ios, @"ios should be present");
    XCTAssertEqualObjects(self.payload.badge, [ios valueForKey:UAChannelBadgeJSONKey], @"badge should be present");
    XCTAssertEqualObjects(self.payload.quietTime, [ios valueForKey:UAChannelQuietTimeJSONKey], @"quiet time should be present");
    XCTAssertEqualObjects(self.payload.quietTimeTimeZone, [ios valueForKey:UAChannelTimeZoneJSONKey], @"timezone should be present");
}

/**
 * Tests that a payload initialized from data is equivalent to standard initialized payload
 */
- (void)testPayloadFromData {
    NSData *payloadData = [self.payload asJSONData];

    UAChannelRegistrationPayload *withData = [UAChannelRegistrationPayload channelRegistrationPayloadWithData:payloadData];
    XCTAssertEqualObjects(withData, self.payload);
}

/**
 * Test when tags are empty or nil
 */
- (void)testAsJsonEmptyTags {
    self.payload.tags = nil;

    NSString *jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    NSDictionary *dict = [NSJSONSerialization objectWithString:jsonString];
    NSDictionary *channel = [dict valueForKey:UAChannelKey];
    XCTAssertNil([channel valueForKey:UAChannelTagsJSONKey], @"tags should be nil");

    // Verify tags is not nil, but an empty nsarray
    self.payload.tags = @[];
    jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    dict = [NSJSONSerialization objectWithString:jsonString];
    channel = [dict valueForKey:UAChannelKey];
    XCTAssertEqualObjects(self.payload.tags, [channel valueForKey:UAChannelTagsJSONKey], @"tags should be nil");
}

/**
 * Test that tags are not sent when setTags is false
 */
- (void)testAsJsonNoTags {
    self.payload.setTags = NO;

    NSString *jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    NSDictionary *dict = [NSJSONSerialization objectWithString:jsonString];
    NSDictionary *channel = [dict valueForKey:UAChannelKey];

    // Verify that tags are not present when setTags is false
    jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    dict = [NSJSONSerialization objectWithString:jsonString];
    channel = [dict valueForKey:UAChannelKey];
    XCTAssertNil([channel valueForKey:UAChannelTagsJSONKey], @"tags should be nil");
}

/**
 * Test that an empty iOS section is not included
 */
- (void)testAsJsonEmptyiOSSection {
    self.payload.badge = nil;
    self.payload.quietTime = nil;
    self.payload.timeZone = nil;

    NSString *jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    NSDictionary *dict = [NSJSONSerialization objectWithString:jsonString];

    XCTAssertNil([dict valueForKey:UAChannelIOSKey], @"iOS section should not be included in the JSON");
}

/**
 * Test that an empty identity hints section is not included
 */
- (void)testAsJsonEmptyIdentityHints {
    self.payload.deviceID = nil;
    self.payload.userID = nil;
    self.payload.accengageDeviceID = nil;

    NSString *jsonString = [[NSString alloc] initWithData:[self.payload asJSONData] encoding:NSUTF8StringEncoding];
    NSDictionary *dict = [NSJSONSerialization objectWithString:jsonString];

    XCTAssertNil([dict valueForKey:UAChannelIdentityHintsKey], @"identity hints section should not be included in the JSON");
}

/**
 * Test isEqualToPayload is equal to its copy
 */
- (void)testisEqualToPayloadCopy {
    UAChannelRegistrationPayload *payloadCopy = [self.payload copy];
    XCTAssertTrue([self.payload isEqualToPayload:payloadCopy], @"A copy should be equal to the original");

    payloadCopy.optedIn = NO;
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.optedIn = self.payload.optedIn;

    payloadCopy.backgroundEnabled = NO;
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.backgroundEnabled = self.payload.backgroundEnabled;

    payloadCopy.pushAddress = @"different-value";
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.pushAddress = self.payload.pushAddress;

    payloadCopy.userID = @"different-value";
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.userID = self.payload.userID;

    payloadCopy.deviceID = @"different-value";
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.deviceID = self.payload.deviceID;

    payloadCopy.badge = [NSNumber numberWithInteger:5];;
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.badge = self.payload.badge;

    payloadCopy.quietTime = [NSDictionary dictionary];
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.quietTime = self.payload.quietTime;

    payloadCopy.timeZone = @"different-value";
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.timeZone = self.payload.timeZone;

    payloadCopy.setTags = NO;
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.setTags = self.payload.setTags;

    payloadCopy.tags = @[@"tagThree", @"tagFour"];;
    XCTAssertFalse([self.payload isEqualToPayload:payloadCopy], @"A payload should not be equal after a modification");
    payloadCopy.tags = self.payload.tags;

    // Make sure its equal again
    XCTAssertTrue([self.payload isEqualToPayload:payloadCopy], @"A copy should be equal to the original");
}

/**
 * Test isEqualToPayload is equal to itself
 */
- (void)testisEqualToPayloadSelf {
    XCTAssertTrue([self.payload isEqualToPayload:self.payload], @"A payload should be equal to itself");
}

/**
 * Test isEqualToPayload is equal to an empty payload
 */
- (void)testisEqualToPayloadEmptyPayload {
    UAChannelRegistrationPayload *emptyPayload = [[UAChannelRegistrationPayload alloc] init];
    XCTAssertFalse([self.payload isEqualToPayload:emptyPayload], @"A payload should not be equal to a different payload");
}

/**
 * Test that payloadDictionary has the full expected payload
 */
- (void)testPayloadDictionaryFullPayload {
    NSDictionary *dict = [[NSDictionary alloc] initWithDictionary:[self.payload payloadDictionary]];

    // identity hints
    NSDictionary *identityHints = [dict valueForKey:UAChannelIdentityHintsKey];
    XCTAssertNotNil(identityHints, @"identity hints should be present");
    XCTAssertEqualObjects(self.payload.userID, [identityHints valueForKey:UAChannelUserIDKey], @"user ID should be present");
    XCTAssertEqualObjects(self.payload.deviceID, [identityHints valueForKey:UAChannelDeviceIDKey], @"device ID should be present");

    // channel specific items
    NSDictionary *channel = [dict valueForKey:UAChannelKey];
    XCTAssertEqualObjects(@"ios", [channel valueForKey:UAChannelDeviceTypeKey], @"device type should be present");
    XCTAssertEqualObjects([NSNumber numberWithBool:self.payload.optedIn], [channel valueForKey:UAChannelOptInKey], @"opt-in should be present");
    XCTAssertEqualObjects(self.payload.pushAddress, [channel valueForKey:UAChannelPushAddressKey], @"push address should be present");
    XCTAssertEqualObjects([NSNumber numberWithBool:self.payload.setTags], [channel valueForKey:UAChannelSetTagsKey], @"set tags should be present");
    XCTAssertEqualObjects(self.payload.tags, [channel valueForKey:UAChannelTagsJSONKey], @"tags should be present");

    // iOS specific items
    NSDictionary *ios = [channel valueForKey:UAChannelIOSKey];
    XCTAssertNotNil(ios, @"ios should be present");
    XCTAssertEqualObjects(self.payload.badge, [ios valueForKey:UAChannelBadgeJSONKey], @"badge should be present");
    XCTAssertEqualObjects(self.payload.quietTime, [ios valueForKey:UAChannelQuietTimeJSONKey], @"quiet time should be present");
    XCTAssertEqualObjects(self.payload.quietTimeTimeZone, [ios valueForKey:UAChannelTimeZoneJSONKey], @"timezone should be present");
}

/**
 * Test payloadDictionary when tags are empty or nil
 */
- (void)testPayloadDictionaryEmptyTags {
    self.payload.tags = nil;
    NSDictionary *dict = [[NSDictionary alloc] initWithDictionary:[self.payload payloadDictionary]];
    NSDictionary *channel = [dict valueForKey:UAChannelKey];
    XCTAssertNil([channel valueForKey:UAChannelTagsJSONKey], @"tags should be nil");

    // Verify tags is not nil, but an empty nsarray
    self.payload.tags = @[];
    dict = [[NSDictionary alloc] initWithDictionary:[self.payload payloadDictionary]];
    channel = [dict valueForKey:UAChannelKey];
    XCTAssertEqualObjects(self.payload.tags, [channel valueForKey:UAChannelTagsJSONKey], @"tags should be nil");
}

/**
 * Test that tags are not sent when setTags is false
 */
- (void)testPayloadDictionaryNoTags {
    self.payload.setTags = NO;

    NSDictionary *dict = [[NSDictionary alloc] initWithDictionary:[self.payload payloadDictionary]];
    NSDictionary *channel = [dict valueForKey:UAChannelKey];

    channel = [dict valueForKey:UAChannelKey];
    XCTAssertNil([channel valueForKey:UAChannelTagsJSONKey], @"tags should be nil");
}

/**
 * Test that an empty iOS section is not included
 */
- (void)testPayloadDictionaryEmptyiOSSection {
    self.payload.badge = nil;
    self.payload.quietTime = nil;
    self.payload.timeZone = nil;

    NSDictionary *dict = [[NSDictionary alloc] initWithDictionary:[self.payload payloadDictionary]];
    XCTAssertNil([dict valueForKey:UAChannelIOSKey], @"iOS section should not be included in the payload");
}

/**
 * Test that an empty identity hints section is not included
 */
- (void)testPayloadDictionaryEmptyIdentityHints {
    self.payload.deviceID = nil;
    self.payload.userID = nil;
    self.payload.accengageDeviceID = nil;

    NSDictionary *dict = [[NSDictionary alloc] initWithDictionary:[self.payload payloadDictionary]];
    XCTAssertNil([dict valueForKey:UAChannelIdentityHintsKey], @"identity hints section should not be included in the payload");
}

- (void)testMinimalUpdatePayloadSameValues {
    UAChannelRegistrationPayload *minPayload = [self.payload minimalUpdatePayloadWithLastPayload:self.payload];
    XCTAssertNil(minPayload.country);
    XCTAssertNil(minPayload.language);
    XCTAssertNil(minPayload.timeZone);
    XCTAssertNil(minPayload.tags);
    XCTAssertNil(minPayload.userID);
    XCTAssertNil(minPayload.deviceID);
    XCTAssertFalse(minPayload.setTags);
    XCTAssertNil(minPayload.locationSettings);
    XCTAssertNil(minPayload.SDKVersion);
    XCTAssertNil(minPayload.appVersion);
    XCTAssertNil(minPayload.deviceModel);
    XCTAssertNil(minPayload.deviceOS);
    XCTAssertNil(minPayload.carrier);

    XCTAssertEqual(self.payload.backgroundEnabled, minPayload.backgroundEnabled);
    XCTAssertEqual(self.payload.optedIn, minPayload.optedIn);
    XCTAssertEqualObjects(self.payload.pushAddress, minPayload.pushAddress);
    XCTAssertEqualObjects(self.payload.quietTimeTimeZone, minPayload.quietTimeTimeZone);
    XCTAssertEqualObjects(self.payload.quietTime, minPayload.quietTime);
    XCTAssertEqualObjects(self.payload.badge, minPayload.badge);
}

- (void)testMinimalUpdatePayloadDifferentValues {
    UAChannelRegistrationPayload *copy = [self.payload copy];
    copy.country = @"country CHANGED";
    copy.timeZone = @"timeZone CHANGED";
    copy.language = @"timeZone CHANGED";
    copy.tags = @[@"tags CHANGED"];
    copy.locationSettings = @(NO);
    copy.SDKVersion = @"SDKVersion CHANGED";
    copy.appVersion = @"appVersion CHANGED";
    copy.deviceModel = @"deviceModel CHANGED";
    copy.deviceOS = @"deviceOS CHANGED";
    copy.carrier = @"carrier CHANGED";

    UAChannelRegistrationPayload *minPayload = [self.payload minimalUpdatePayloadWithLastPayload:copy];
    XCTAssertEqualObjects(self.payload.country, minPayload.country);
    XCTAssertEqualObjects(self.payload.timeZone, minPayload.timeZone);
    XCTAssertEqualObjects(self.payload.language, minPayload.language);
    XCTAssertEqualObjects(self.payload.tags, minPayload.tags);
    XCTAssertEqualObjects(self.payload.locationSettings, minPayload.locationSettings);
    XCTAssertEqualObjects(self.payload.SDKVersion, minPayload.SDKVersion);
    XCTAssertEqualObjects(self.payload.appVersion, minPayload.appVersion);
    XCTAssertEqualObjects(self.payload.deviceModel, minPayload.deviceModel);
    XCTAssertEqualObjects(self.payload.deviceOS, minPayload.deviceOS);
    XCTAssertEqualObjects(self.payload.carrier, minPayload.carrier);

    XCTAssertNil(minPayload.userID);
    XCTAssertNil(minPayload.deviceID);
    XCTAssertTrue(minPayload.setTags);

    XCTAssertEqual(self.payload.backgroundEnabled, minPayload.backgroundEnabled);
    XCTAssertEqual(self.payload.optedIn, minPayload.optedIn);
    XCTAssertEqualObjects(self.payload.pushAddress, minPayload.pushAddress);
    XCTAssertEqualObjects(self.payload.quietTimeTimeZone, minPayload.quietTimeTimeZone);
    XCTAssertEqualObjects(self.payload.quietTime, minPayload.quietTime);
    XCTAssertEqualObjects(self.payload.badge, minPayload.badge);
}

- (void)testMinimalUpdatePayloadDifferentNamedUserId {
    UAChannelRegistrationPayload *thisPayload = [self.payload copy];
    thisPayload.namedUserId = @"differentFakeNamedUserID";

    UAChannelRegistrationPayload *minPayload = [thisPayload minimalUpdatePayloadWithLastPayload:self.payload];
    XCTAssertEqualObjects(thisPayload.country, minPayload.country);
    XCTAssertEqualObjects(thisPayload.timeZone, minPayload.timeZone);
    XCTAssertEqualObjects(thisPayload.language, minPayload.language);
    XCTAssertNil(minPayload.tags);
    XCTAssertEqualObjects(thisPayload.locationSettings, minPayload.locationSettings);
    XCTAssertEqualObjects(thisPayload.SDKVersion, minPayload.SDKVersion);
    XCTAssertEqualObjects(thisPayload.appVersion, minPayload.appVersion);
    XCTAssertEqualObjects(thisPayload.deviceModel, minPayload.deviceModel);
    XCTAssertEqualObjects(thisPayload.deviceOS, minPayload.deviceOS);
    XCTAssertEqualObjects(thisPayload.carrier, minPayload.carrier);

    XCTAssertNil(minPayload.userID);
    XCTAssertNil(minPayload.deviceID);
    XCTAssertNil(minPayload.accengageDeviceID);
    XCTAssertFalse(minPayload.setTags);

    XCTAssertEqual(thisPayload.backgroundEnabled, minPayload.backgroundEnabled);
    XCTAssertEqual(thisPayload.optedIn, minPayload.optedIn);
    XCTAssertEqualObjects(thisPayload.pushAddress, minPayload.pushAddress);
    XCTAssertEqualObjects(thisPayload.quietTimeTimeZone, minPayload.quietTimeTimeZone);
    XCTAssertEqualObjects(thisPayload.quietTime, minPayload.quietTime);
    XCTAssertEqualObjects(thisPayload.badge, minPayload.badge);
}

- (void)testMinimalUpdatePayloadChangeToNilNamedUserId {
    UAChannelRegistrationPayload *thisPayload = [self.payload copy];
    thisPayload.namedUserId = nil;

    UAChannelRegistrationPayload *minPayload = [thisPayload minimalUpdatePayloadWithLastPayload:self.payload];
    XCTAssertNil(minPayload.country);
    XCTAssertNil(minPayload.timeZone);
    XCTAssertNil(minPayload.language);
    XCTAssertNil(minPayload.tags);
    XCTAssertNil(minPayload.locationSettings);
    XCTAssertNil(minPayload.SDKVersion);
    XCTAssertNil(minPayload.appVersion);
    XCTAssertNil(minPayload.deviceModel);
    XCTAssertNil(minPayload.deviceOS);
    XCTAssertNil(minPayload.carrier);

    XCTAssertNil(minPayload.userID);
    XCTAssertNil(minPayload.deviceID);
    XCTAssertNil(minPayload.accengageDeviceID);
    XCTAssertFalse(minPayload.setTags);

    XCTAssertEqual(thisPayload.backgroundEnabled, minPayload.backgroundEnabled);
    XCTAssertEqual(thisPayload.optedIn, minPayload.optedIn);
    XCTAssertEqualObjects(thisPayload.pushAddress, minPayload.pushAddress);
    XCTAssertEqualObjects(thisPayload.quietTimeTimeZone, minPayload.quietTimeTimeZone);
    XCTAssertEqualObjects(thisPayload.quietTime, minPayload.quietTime);
    XCTAssertEqualObjects(thisPayload.badge, minPayload.badge);
}

- (void)testMinimalUpdatePayloadNilValues {
    UAChannelRegistrationPayload *copy = [self.payload copy];
    copy.country = nil;
    copy.timeZone = nil;
    copy.language = nil;
    copy.tags = nil;
    copy.locationSettings = nil;
    copy.SDKVersion = nil;
    copy.appVersion = nil;
    copy.deviceModel = nil;
    copy.deviceOS = nil;
    copy.carrier = nil;

    UAChannelRegistrationPayload *minPayload = [self.payload minimalUpdatePayloadWithLastPayload:copy];
    XCTAssertEqualObjects(self.payload.country, minPayload.country);
    XCTAssertEqualObjects(self.payload.timeZone, minPayload.timeZone);
    XCTAssertEqualObjects(self.payload.language, minPayload.language);
    XCTAssertEqualObjects(self.payload.tags, minPayload.tags);
    XCTAssertEqualObjects(self.payload.locationSettings, minPayload.locationSettings);
    XCTAssertEqualObjects(self.payload.SDKVersion, minPayload.SDKVersion);
    XCTAssertEqualObjects(self.payload.appVersion, minPayload.appVersion);
    XCTAssertEqualObjects(self.payload.deviceModel, minPayload.deviceModel);
    XCTAssertEqualObjects(self.payload.deviceOS, minPayload.deviceOS);
    XCTAssertEqualObjects(self.payload.carrier, minPayload.carrier);

    XCTAssertNil(minPayload.userID);
    XCTAssertNil(minPayload.deviceID);
    XCTAssertTrue(minPayload.setTags);

    XCTAssertEqual(self.payload.backgroundEnabled, minPayload.backgroundEnabled);
    XCTAssertEqual(self.payload.optedIn, minPayload.optedIn);
    XCTAssertEqualObjects(self.payload.pushAddress, minPayload.pushAddress);
    XCTAssertEqualObjects(self.payload.quietTimeTimeZone, minPayload.quietTimeTimeZone);
    XCTAssertEqualObjects(self.payload.quietTime, minPayload.quietTime);
    XCTAssertEqualObjects(self.payload.badge, minPayload.badge);
}

#pragma mark -
#pragma mark Helpers

- (NSMutableDictionary *)buildQuietTimeWithStartDate:(NSDate *)startDate withEndDate:(NSDate *)endDate {

    NSCalendar *cal = [[NSCalendar alloc] initWithCalendarIdentifier:NSCalendarIdentifierGregorian];
    NSString *fromStr = [NSString stringWithFormat:@"%ld:%02ld",
                         (long)[cal components:NSCalendarUnitHour fromDate:startDate].hour,
                         (long)[cal components:NSCalendarUnitMinute fromDate:startDate].minute];

    NSString *toStr = [NSString stringWithFormat:@"%ld:%02ld",
                       (long)[cal components:NSCalendarUnitHour fromDate:endDate].hour,
                       (long)[cal components:NSCalendarUnitMinute fromDate:endDate].minute];

    return [NSMutableDictionary dictionaryWithObjectsAndKeys:
            fromStr, UAPushQuietTimeStartKey,
            toStr, UAPushQuietTimeEndKey, nil];
    
}

@end
